@@ -1,35 +1,53 @@
-# Copyright (c) 2003-2013 LOGILAB S.A. (Paris, FRANCE).
-# http://www.logilab.fr/ -- mailto:contact@logilab.fr
-#
-# This program is free software; you can redistribute it and/or modify it under
-# the terms of the GNU General Public License as published by the Free Software
-# Foundation; either version 2 of the License, or (at your option) any later
-# version.
-#
-# This program is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License along with
-# this program; if not, write to the Free Software Foundation, Inc.,
-# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+# -*- coding: utf-8 -*-
+# Copyright (c) 2006-2015 LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
+# Copyright (c) 2012-2014 Google, Inc.
+# Copyright (c) 2013 buck@yelp.com <buck@yelp.com>
+# Copyright (c) 2014-2017 Claudiu Popa <pcmanticore@gmail.com>
+# Copyright (c) 2014 Brett Cannon <brett@python.org>
+# Copyright (c) 2014 Arun Persaud <arun@nubati.net>
+# Copyright (c) 2015-2016 Moises Lopez <moylop260@vauxoo.com>
+# Copyright (c) 2015 Dmitry Pribysh <dmand@yandex.ru>
+# Copyright (c) 2015 Cezar <celnazli@bitdefender.com>
+# Copyright (c) 2015 Florian Bruhin <me@the-compiler.org>
+# Copyright (c) 2015 Noam Yorav-Raphael <noamraph@gmail.com>
+# Copyright (c) 2015 James Morgensen <james.morgensen@gmail.com>
+# Copyright (c) 2015 Ionel Cristian Maries <contact@ionelmc.ro>
+# Copyright (c) 2016 Jared Garst <cultofjared@gmail.com>
+# Copyright (c) 2016 Maik Röder <maikroeder@gmail.com>
+# Copyright (c) 2016 Glenn Matthews <glenn@e-dad.net>
+# Copyright (c) 2016 Ashley Whetter <ashley@awhetter.co.uk>
+# Copyright (c) 2017 hippo91 <guillaume.peillex@gmail.com>
+# Copyright (c) 2017 Michka Popoff <michkapopoff@gmail.com>
+# Copyright (c) 2017 Łukasz Rogalski <rogalski.91@gmail.com>
+# Copyright (c) 2017 Erik Wright <erik.wright@shopify.com>
+
+# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
+# For details: https://github.com/PyCQA/pylint/blob/master/COPYING
+
 """imports checkers for Python code"""
 
 import collections
+from distutils import sysconfig
 import os
 import sys
+import copy
 
 import six
 
 import astroid
 from astroid import are_exclusive
-from astroid.modutils import (EXT_LIB_DIR, get_module_part, is_standard_module,
-                              file_from_modpath)
+from astroid.modutils import (get_module_part, is_standard_module)
+import isort
 
 from pylint.interfaces import IAstroidChecker
-from pylint.utils import EmptyReport, get_global_option
+from pylint.utils import get_global_option
+from pylint.exceptions import EmptyReportError
 from pylint.checkers import BaseChecker
-from pylint.checkers.utils import check_messages, node_ignores_exception
+from pylint.checkers.utils import (
+    check_messages,
+    node_ignores_exception,
+    is_from_fallback_block
+)
 from pylint.graph import get_cycles, DotBackend
 from pylint.reporters.ureports.nodes import VerbatimText, Paragraph
 
@@ -63,7 +81,7 @@ def _get_import_name(importnode, modname):
     return modname
 
 
-def _get_first_import(node, context, name, base, level):
+def _get_first_import(node, context, name, base, level, alias):
     """return the node where [base.]<name> is imported or None if not found
     """
     fullname = '%s.%s' % (base, name) if base else name
@@ -80,13 +98,27 @@ def _get_first_import(node, context, name, base, level):
                 found = True
                 break
         elif isinstance(first, astroid.ImportFrom):
-            if level == first.level and any(
-                    fullname == '%s.%s' % (first.modname, iname[0])
-                    for iname in first.names):
-                found = True
-                break
+            if level == first.level:
+                for imported_name, imported_alias in first.names:
+                    if fullname == '%s.%s' % (first.modname, imported_name):
+                        found = True
+                        break
+                    if name != '*' and name == imported_name and not (alias or imported_alias):
+                        found = True
+                        break
+                if found:
+                    break
     if found and not are_exclusive(first, node):
         return first
+    return None
+
+
+def _ignore_import_failure(node, modname, ignored_modules):
+    for submodule in _qualified_names(modname):
+        if submodule in ignored_modules:
+            return True
+
+    return node_ignores_exception(node, ImportError)
 
 # utilities to represents import dependencies as tree and dot graph ###########
 
@@ -111,7 +143,7 @@ def _repr_tree_defs(data, indent_str=None):
         if not files:
             files = ''
         else:
-            files = '(%s)' % ','.join(files)
+            files = '(%s)' % ','.join(sorted(files))
         if indent_str is None:
             lines.append('%s %s' % (mod, files))
             sub_indent_str = '  '
@@ -135,10 +167,10 @@ def _dependencies_graph(filename, dep_info):
     for modname, dependencies in sorted(six.iteritems(dep_info)):
         done[modname] = 1
         printer.emit_node(modname)
-        for modname in dependencies:
-            if modname not in done:
-                done[modname] = 1
-                printer.emit_node(modname)
+        for depmodname in dependencies:
+            if depmodname not in done:
+                done[depmodname] = 1
+                printer.emit_node(depmodname)
     for depmodname, dependencies in sorted(six.iteritems(dep_info)):
         for modname in dependencies:
             printer.emit_edge(modname, depmodname)
@@ -161,6 +193,10 @@ MSGS = {
               'import-error',
               'Used when pylint has been unable to import a module.',
               {'old_names': [('F0401', 'import-error')]}),
+    'E0402': ('Attempted relative import beyond top-level package',
+              'relative-beyond-top-level',
+              'Used when a relative import tries to access too many levels '
+              'in the current package.'),
     'R0401': ('Cyclic import (%s)',
               'cyclic-import',
               'Used when a cyclic import between two or more modules is \
@@ -193,7 +229,7 @@ MSGS = {
               'multiple-imports',
               'Used when import statement importing multiple modules is '
               'detected.'),
-    'C0411': ('%s comes before %s',
+    'C0411': ('%s should be placed before %s',
               'wrong-import-order',
               'Used when PEP8 import order is not respected (standard imports '
               'first, then third-party libraries, then local imports)'),
@@ -206,6 +242,11 @@ MSGS = {
               'Used when code and imports are mixed'),
     }
 
+
+DEFAULT_STANDARD_LIBRARY = ()
+DEFAULT_KNOWN_THIRD_PARTY = ('enchant',)
+
+
 class ImportsChecker(BaseChecker):
     """checks for
     * external modules dependencies
@@ -222,38 +263,68 @@ class ImportsChecker(BaseChecker):
 
     if six.PY2:
         deprecated_modules = ('regsub', 'TERMIOS', 'Bastion', 'rexec')
-    else:
+    elif sys.version_info < (3, 5):
         deprecated_modules = ('optparse', )
+    else:
+        deprecated_modules = ('optparse', 'tkinter.tix')
     options = (('deprecated-modules',
                 {'default' : deprecated_modules,
                  'type' : 'csv',
                  'metavar' : '<modules>',
-                 'help' : 'Deprecated modules which should not be used, \
-separated by a comma'}
+                 'help' : 'Deprecated modules which should not be used,'
+                          ' separated by a comma'}
                ),
                ('import-graph',
                 {'default' : '',
                  'type' : 'string',
                  'metavar' : '<file.dot>',
-                 'help' : 'Create a graph of every (i.e. internal and \
-external) dependencies in the given file (report RP0402 must not be disabled)'}
+                 'help' : 'Create a graph of every (i.e. internal and'
+                          ' external) dependencies in the given file'
+                          ' (report RP0402 must not be disabled)'}
                ),
                ('ext-import-graph',
                 {'default' : '',
                  'type' : 'string',
                  'metavar' : '<file.dot>',
-                 'help' : 'Create a graph of external dependencies in the \
-given file (report RP0402 must not be disabled)'}
+                 'help' : 'Create a graph of external dependencies in the'
+                          ' given file (report RP0402 must not be disabled)'}
                ),
                ('int-import-graph',
                 {'default' : '',
                  'type' : 'string',
                  'metavar' : '<file.dot>',
-                 'help' : 'Create a graph of internal dependencies in the \
-given file (report RP0402 must not be disabled)'}
+                 'help' : 'Create a graph of internal dependencies in the'
+                          ' given file (report RP0402 must not be disabled)'}
                ),
+               ('known-standard-library',
+                {'default': DEFAULT_STANDARD_LIBRARY,
+                 'type': 'csv',
+                 'metavar': '<modules>',
+                 'help': 'Force import order to recognize a module as part of'
+                         ' the standard compatibility libraries.'}
+               ),
+               ('known-third-party',
+                {'default': DEFAULT_KNOWN_THIRD_PARTY,
+                 'type': 'csv',
+                 'metavar': '<modules>',
+                 'help': 'Force import order to recognize a module as part of'
+                         ' a third party library.'}
+               ),
+               ('analyse-fallback-blocks',
+                {'default': False,
+                 'type': 'yn',
+                 'metavar': '<y_or_n>',
+                 'help': 'Analyse import fallback blocks. This can be used to '
+                         'support both Python 2 and 3 compatible code, which means that '
+                         'the block might have code that exists only in one or another '
+                         'interpreter, leading to false positives when analysed.'},
+               ),
+               ('allow-wildcard-with-all',
+                {'default': False,
+                 'type': 'yn',
+                 'metavar': '<y_or_n>',
+                 'help': 'Allow wildcard imports from modules that define __all__.'}),
               )
-    ext_lib_dir = os.path.normcase(os.path.abspath(EXT_LIB_DIR))
 
     def __init__(self, linter=None):
         BaseChecker.__init__(self, linter)
@@ -268,25 +339,54 @@ given file (report RP0402 must not be disabled)'}
                          self._report_dependencies_graph),
                        )
 
+        self._site_packages = self._compute_site_packages()
+
+    @staticmethod
+    def _compute_site_packages():
+        def _normalized_path(path):
+            return os.path.normcase(os.path.abspath(path))
+
+        paths = set()
+        real_prefix = getattr(sys, 'real_prefix', None)
+        for prefix in filter(None, (real_prefix, sys.prefix)):
+            path = sysconfig.get_python_lib(prefix=prefix)
+            path = _normalized_path(path)
+            paths.add(path)
+
+        # Handle Debian's derivatives /usr/local.
+        if os.path.isfile("/etc/debian_version"):
+            for prefix in filter(None, (real_prefix, sys.prefix)):
+                libpython = os.path.join(prefix, "local", "lib",
+                                         "python" + sysconfig.get_python_version(),
+                                         "dist-packages")
+                paths.add(libpython)
+        return paths
+
     def open(self):
         """called before visiting project (i.e set of modules)"""
         self.linter.add_stats(dependencies={})
         self.linter.add_stats(cycles=[])
         self.stats = self.linter.stats
         self.import_graph = collections.defaultdict(set)
+        self._excluded_edges = collections.defaultdict(set)
         self._ignored_modules = get_global_option(
             self, 'ignored-modules', default=[])
 
+    def _import_graph_without_ignored_edges(self):
+        filtered_graph = copy.deepcopy(self.import_graph)
+        for node in filtered_graph:
+            filtered_graph[node].difference_update(self._excluded_edges[node])
+        return filtered_graph
+
     def close(self):
         """called before visiting project (i.e set of modules)"""
-        # don't try to compute cycles if the associated message is disabled
         if self.linter.is_message_enabled('cyclic-import'):
-            vertices = list(self.import_graph)
-            for cycle in get_cycles(self.import_graph, vertices=vertices):
+            graph = self._import_graph_without_ignored_edges()
+            vertices = list(graph)
+            for cycle in get_cycles(graph, vertices=vertices):
                 self.add_message('cyclic-import', args=' -> '.join(cycle))
 
-    @check_messages('wrong-import-position', 'multiple-imports',
-                    'relative-import', 'reimported')
+    @check_messages(*MSGS.keys())
     def visit_import(self, node):
         """triggered when an import statement is seen"""
         self._check_reimport(node)
@@ -298,42 +398,46 @@ given file (report RP0402 must not be disabled)'}
 
         for name in names:
             self._check_deprecated_module(node, name)
-            importedmodnode = self.get_imported_module(node, name)
-            if isinstance(node.scope(), astroid.Module):
+            imported_module = self._get_imported_module(node, name)
+            if isinstance(node.parent, astroid.Module):
+                # Allow imports nested
                 self._check_position(node)
-                self._record_import(node, importedmodnode)
+            if isinstance(node.scope(), astroid.Module):
+                self._record_import(node, imported_module)
 
-            if importedmodnode is None:
+            if imported_module is None:
                 continue
 
-            self._check_relative_import(modnode, node, importedmodnode, name)
-            self._add_imported_module(node, importedmodnode.name)
+            self._check_relative_import(modnode, node, imported_module, name)
+            self._add_imported_module(node, imported_module.name)
 
     @check_messages(*(MSGS.keys()))
     def visit_importfrom(self, node):
         """triggered when a from statement is seen"""
         basename = node.modname
+        imported_module = self._get_imported_module(node, basename)
+
         self._check_misplaced_future(node)
         self._check_deprecated_module(node, basename)
-        self._check_wildcard_imports(node)
+        self._check_wildcard_imports(node, imported_module)
         self._check_same_line_imports(node)
         self._check_reimport(node, basename=basename, level=node.level)
 
-        modnode = node.root()
-        importedmodnode = self.get_imported_module(node, basename)
-        if isinstance(node.scope(), astroid.Module):
+        if isinstance(node.parent, astroid.Module):
+            # Allow imports nested
             self._check_position(node)
-            self._record_import(node, importedmodnode)
-        if importedmodnode is None:
+        if isinstance(node.scope(), astroid.Module):
+            self._record_import(node, imported_module)
+        if imported_module is None:
             return
-        self._check_relative_import(modnode, node, importedmodnode, basename)
+        modnode = node.root()
+        self._check_relative_import(modnode, node, imported_module, basename)
 
         for name, _ in node.names:
             if name != '*':
-                self._add_imported_module(node, '%s.%s' % (importedmodnode.name, name))
+                self._add_imported_module(node, '%s.%s' % (imported_module.name, name))
 
-    @check_messages('wrong-import-order', 'ungrouped-imports',
-                    'wrong-import-position')
+    @check_messages(*(MSGS.keys()))
     def leave_module(self, node):
         # Check imports are grouped by category (standard, 3rd party, local)
         std_imports, ext_imports, loc_imports = self._check_imports_order(node)
@@ -342,6 +446,8 @@ given file (report RP0402 must not be disabled)'}
         met = set()
         current_package = None
         for import_node, import_name in std_imports + ext_imports + loc_imports:
+            if not self.linter.is_message_enabled('ungrouped-imports', import_node.fromlineno):
+                continue
             package, _, _ = import_name.partition('.')
             if current_package and current_package != package and package in met:
                 self.add_message('ungrouped-imports', node=import_node,
@@ -352,7 +458,9 @@ given file (report RP0402 must not be disabled)'}
         self._imports_stack = []
         self._first_non_import_node = None
 
-    def visit_if(self, node):
+    def compute_first_non_import_node(self, node):
+        if not self.linter.is_message_enabled('wrong-import-position', node.fromlineno):
+            return
         # if the node does not contain an import instruction, and if it is the
         # first node of the module, keep a track of it (all the import positions
         # of the module will be compared to the position of this first
@@ -361,14 +469,30 @@ given file (report RP0402 must not be disabled)'}
             return
         if not isinstance(node.parent, astroid.Module):
             return
-        if any(node.nodes_of_class((astroid.Import, astroid.ImportFrom))):
+        nested_allowed = [astroid.TryExcept, astroid.TryFinally]
+        is_nested_allowed = [
+            allowed for allowed in nested_allowed if isinstance(node, allowed)]
+        if is_nested_allowed and \
+                any(node.nodes_of_class((astroid.Import, astroid.ImportFrom))):
             return
+        if isinstance(node, astroid.Assign):
+            # Add compatibility for module level dunder names
+            # https://www.python.org/dev/peps/pep-0008/#module-level-dunder-names
+            valid_targets = [
+                isinstance(target, astroid.AssignName) and
+                target.name.startswith('__') and target.name.endswith('__')
+                for target in node.targets]
+            if all(valid_targets):
+                return
         self._first_non_import_node = node
 
-    visit_tryfinally = visit_tryexcept = visit_assignattr = visit_assign \
-            = visit_ifexp = visit_comprehension = visit_if
+    visit_tryfinally = visit_tryexcept = visit_assignattr = visit_assign = \
+        visit_ifexp = visit_comprehension = visit_expr = visit_if = \
+        compute_first_non_import_node
 
     def visit_functiondef(self, node):
+        if not self.linter.is_message_enabled('wrong-import-position', node.fromlineno):
+            return
         # If it is the first non import instruction of the module, record it.
         if self._first_non_import_node:
             return
@@ -426,69 +550,109 @@ given file (report RP0402 must not be disabled)'}
         """Record the package `node` imports from"""
         importedname = importedmodnode.name if importedmodnode else None
         if not importedname:
-            importedname = node.names[0][0].split('.')[0]
+            if isinstance(node, astroid.ImportFrom):
+                importedname = node.modname
+            else:
+                importedname = node.names[0][0].split('.')[0]
+        if isinstance(node, astroid.ImportFrom) and (node.level or 0) >= 1:
+            # We need the impotedname with first point to detect local package
+            # Example of node:
+            #  'from .my_package1 import MyClass1'
+            #  the output should be '.my_package1' instead of 'my_package1'
+            # Example of node:
+            #  'from . import my_package2'
+            #  the output should be '.my_package2' instead of '{pyfile}'
+            importedname = '.' + importedname
         self._imports_stack.append((node, importedname))
 
-    def _check_imports_order(self, node):
+    @staticmethod
+    def _is_fallback_import(node, imports):
+        imports = [import_node for (import_node, _) in imports]
+        return any(astroid.are_exclusive(import_node, node)
+                   for import_node in imports)
+
+    def _check_imports_order(self, _module_node):
         """Checks imports of module `node` are grouped by category
 
         Imports must follow this order: standard, 3rd party, local
         """
-        extern_imports = []
-        local_imports = []
         std_imports = []
-        stdlib_paths = [sys.prefix, self.ext_lib_dir]
-        real_prefix = getattr(sys, 'real_prefix', None)
-        if real_prefix is not None:
-            stdlib_paths.append(real_prefix)
-
+        third_party_imports = []
+        first_party_imports = []
+        # need of a list that holds third or first party ordered import
+        external_imports = []
+        local_imports = []
+        third_party_not_ignored = []
+        first_party_not_ignored = []
+        local_not_ignored = []
+        isort_obj = isort.SortImports(
+            file_contents='', known_third_party=self.config.known_third_party,
+            known_standard_library=self.config.known_standard_library,
+        )
         for node, modname in self._imports_stack:
-            package = modname.split('.')[0]
-            if is_standard_module(modname):
-                std_imports.append((node, package))
-                wrong_import = extern_imports or local_imports
-                if not wrong_import:
-                    continue
-                self.add_message('wrong-import-order', node=node,
-                                 args=('standard import "%s"' % node.as_string(),
-                                       '"%s"' % wrong_import[0][0].as_string()))
+            if modname.startswith('.'):
+                package = '.' + modname.split('.')[1]
             else:
-                try:
-                    filename = file_from_modpath([package])
-                except ImportError:
-                    extern_imports.append((node, package))
+                package = modname.split('.')[0]
+            nested = not isinstance(node.parent, astroid.Module)
+            ignore_for_import_order = not self.linter.is_message_enabled('wrong-import-order',
+                                                                         node.fromlineno)
+            import_category = isort_obj.place_module(package)
+            node_and_package_import = (node, package)
+            if import_category in ('FUTURE', 'STDLIB'):
+                std_imports.append(node_and_package_import)
+                wrong_import = (third_party_not_ignored or first_party_not_ignored
+                                or local_not_ignored)
+                if self._is_fallback_import(node, wrong_import):
                     continue
-                if not filename:
-                    extern_imports.append((node, package))
-                    continue
-                filename = os.path.normcase(os.path.abspath(filename))
-                if not any(filename.startswith(path) for path in stdlib_paths):
-                    local_imports.append((node, package))
-                    continue
-                extern_imports.append((node, package))
-                if not local_imports:
-                    continue
-                self.add_message('wrong-import-order', node=node,
-                                 args=('external import "%s"' % node.as_string(),
-                                       '"%s"' % local_imports[0][0].as_string()))
-        return std_imports, extern_imports, local_imports
+                if wrong_import and not nested:
+                    self.add_message('wrong-import-order', node=node,
+                                     args=('standard import "%s"' % node.as_string(),
+                                           '"%s"' % wrong_import[0][0].as_string()))
+            elif import_category == 'THIRDPARTY':
+                third_party_imports.append(node_and_package_import)
+                external_imports.append(node_and_package_import)
+                if not nested and not ignore_for_import_order:
+                    third_party_not_ignored.append(node_and_package_import)
+                wrong_import = first_party_not_ignored or local_not_ignored
+                if wrong_import and not nested:
+                    self.add_message('wrong-import-order', node=node,
+                                     args=('third party import "%s"' % node.as_string(),
+                                           '"%s"' % wrong_import[0][0].as_string()))
+            elif import_category == 'FIRSTPARTY':
+                first_party_imports.append(node_and_package_import)
+                external_imports.append(node_and_package_import)
+                if not nested and not ignore_for_import_order:
+                    first_party_not_ignored.append(node_and_package_import)
+                wrong_import = local_not_ignored
+                if wrong_import and not nested:
+                    self.add_message('wrong-import-order', node=node,
+                                     args=('first party import "%s"' % node.as_string(),
+                                           '"%s"' % wrong_import[0][0].as_string()))
+            elif import_category == 'LOCALFOLDER':
+                local_imports.append((node, package))
+                if not nested and not ignore_for_import_order:
+                    local_not_ignored.append((node, package))
+        return std_imports, external_imports, local_imports
 
-    def get_imported_module(self, importnode, modname):
+    def _get_imported_module(self, importnode, modname):
         try:
             return importnode.do_import_module(modname)
-        except astroid.InferenceError as ex:
+        except astroid.TooManyLevelsError:
+            if _ignore_import_failure(importnode, modname, self._ignored_modules):
+                return None
+
+            self.add_message('relative-beyond-top-level', node=importnode)
+
+        except astroid.AstroidBuildingException:
+            if _ignore_import_failure(importnode, modname, self._ignored_modules):
+                return None
+            if not self.config.analyse_fallback_blocks and is_from_fallback_block(importnode):
+                return None
+
             dotted_modname = _get_import_name(importnode, modname)
-            if str(ex) != modname:
-                args = '%r (%s)' % (dotted_modname, ex)
-            else:
-                args = repr(dotted_modname)
-
-            for submodule in _qualified_names(modname):
-                if submodule in self._ignored_modules:
-                    return None
-
-            if not node_ignores_exception(importnode, ImportError):
-                self.add_message("import-error", args=args, node=importnode)
+            self.add_message('import-error', args=repr(dotted_modname),
+                             node=importnode)
 
     def _check_relative_import(self, modnode, importnode, importedmodnode,
                                importedasname):
@@ -496,7 +660,7 @@ given file (report RP0402 must not be disabled)'}
         the imported module name.
         """
         if not self.linter.is_message_enabled('relative-import'):
-            return
+            return None
         if importedmodnode.file is None:
             return False # built-in module
         if modnode is importedmodnode:
@@ -508,6 +672,8 @@ given file (report RP0402 must not be disabled)'}
             self.add_message('relative-import',
                              args=(importedasname, importedmodnode.name),
                              node=importnode)
+            return None
+        return None
 
     def _add_imported_module(self, node, importedmodname):
         """notify an imported module, used to analyze dependencies"""
@@ -538,10 +704,11 @@ given file (report RP0402 must not be disabled)'}
                 importedmodname, set())
             if context_name not in importedmodnames:
                 importedmodnames.add(context_name)
+
             # update import graph
-            mgraph = self.import_graph[context_name]
-            if importedmodname not in mgraph:
-                mgraph.add(importedmodname)
+            self.import_graph[context_name].add(importedmodname)
+            if not self.linter.is_message_enabled('cyclic-import', line=node.lineno):
+                self._excluded_edges[context_name].add(importedmodname)
 
     def _check_deprecated_module(self, node, mod_path):
         """check if the module is deprecated"""
@@ -560,28 +727,31 @@ given file (report RP0402 must not be disabled)'}
         if root is not frame:
             contexts.append((root, None))
 
-        for context, level in contexts:
-            for name, _ in node.names:
-                first = _get_first_import(node, context, name, basename, level)
+        for known_context, known_level in contexts:
+            for name, alias in node.names:
+                first = _get_first_import(
+                    node, known_context,
+                    name, basename,
+                    known_level, alias)
                 if first is not None:
                     self.add_message('reimported', node=node,
                                      args=(name, first.fromlineno))
 
-    def _report_external_dependencies(self, sect, _, dummy):
+    def _report_external_dependencies(self, sect, _, _dummy):
         """return a verbatim layout for displaying dependencies"""
         dep_info = _make_tree_defs(six.iteritems(self._external_dependencies_info()))
         if not dep_info:
-            raise EmptyReport()
+            raise EmptyReportError()
         tree_str = _repr_tree_defs(dep_info)
         sect.append(VerbatimText(tree_str))
 
-    def _report_dependencies_graph(self, sect, _, dummy):
+    def _report_dependencies_graph(self, sect, _, _dummy):
         """write dependencies as a dot (graphviz) file"""
         dep_info = self.stats['dependencies']
         if not dep_info or not (self.config.import_graph
                                 or self.config.ext_import_graph
                                 or self.config.int_import_graph):
-            raise EmptyReport()
+            raise EmptyReportError()
         filename = self.config.import_graph
         if filename:
             _make_graph(filename, dep_info, sect, '')
@@ -618,11 +788,19 @@ given file (report RP0402 must not be disabled)'}
                     result[importee] = importers
         return self.__int_dep_info
 
-    def _check_wildcard_imports(self, node):
+    def _check_wildcard_imports(self, node, imported_module):
+        wildcard_import_is_allowed = (
+            self._wildcard_import_is_allowed(imported_module)
+        )
         for name, _ in node.names:
-            if name == '*':
+            if name == '*' and not wildcard_import_is_allowed:
                 self.add_message('wildcard-import', args=node.modname, node=node)
 
+    def _wildcard_import_is_allowed(self, imported_module):
+        return (self.config.allow_wildcard_with_all
+                and imported_module is not None
+                and '__all__' in imported_module.locals)
+
 
 def register(linter):
     """required method to auto register this checker """
